Introduce the vss task.
[paraslash.git] / sched.c
diff --git a/sched.c b/sched.c
index e122903e8e46823ed2a4b987cf3264782ca3040e..02793124f38b2c8a185dbe6fa7e1b27fd5306ab2 100644 (file)
--- a/sched.c
+++ b/sched.c
@@ -1,24 +1,15 @@
 /*
- * Copyright (C) 2006 Andre Noll <maan@systemlinux.org>
+ * Copyright (C) 2006-2008 Andre Noll <maan@systemlinux.org>
  *
- *     This program is free software; you can redistribute it and/or modify
- *     it under the terms of the GNU General Public License as published by
- *     the Free Software Foundation; either version 2 of the License, or
- *     (at your option) any later version.
- *
- *     This program is distributed in the hope that it will be useful,
- *     but WITHOUT ANY WARRANTY; without even the implied warranty of
- *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *     GNU General Public License for more details.
- *
- *     You should have received a copy of the GNU General Public License
- *     along with this program; if not, write to the Free Software
- *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
+ * Licensed under the GPL v2. For licencing details see COPYING.
  */
 
-/** \file sched.c paraslash's scheduling functions */
+/** \file sched.c Paraslash's scheduling functions. */
 
+#include <dirent.h> /* readdir() */
+#include <assert.h>
 #include <sys/time.h>
+
 #include "para.h"
 #include "ipc.h"
 #include "fd.h"
@@ -33,19 +24,46 @@ static int initialized;
 static struct timeval now_struct;
 struct timeval *now = &now_struct;
 
+/**
+ * Remove a task from the scheduler.
+ *
+ * \param t The task to remove.
+ *
+ * If the pre_select pointer of \a t is not \p NULL, it is removed from
+ * the pre_select list of the scheduler. Same goes for \a post_select.
+ */
+static void unregister_task(struct task *t)
+{
+       if (!initialized)
+               return;
+       PARA_INFO_LOG("unregistering %s (%s)\n", t->status,
+               t->error <0? para_strerror(-t->error) : "shutdown");
+       if (t->pre_select)
+               list_del(&t->pre_select_node);
+       if (t->post_select)
+               list_del(&t->post_select_node);
+       t->error = -E_TASK_UNREGISTERED;
+};
+
+
 static void sched_preselect(struct sched *s)
 {
        struct task *t, *tmp;
-again:
        list_for_each_entry_safe(t, tmp, &pre_select_list, pre_select_node) {
-               t->pre_select(s, t);
+               if (t->error >= 0)
+                       t->pre_select(s, t);
 //             PARA_INFO_LOG("%s \n", t->status);
-               if (t->ret > 0)
-                       continue;
-               if (!t->event_handler)
+               if (t->error >= 0)
                        continue;
-               t->event_handler(t);
-               goto again;
+               /*
+                * We have to check whether the list is empty because the call
+                * to ->pre_select() might have called sched_shutdown(). In
+                * this case t has been unregistered already, so we must not
+                * unregister it again.
+                */
+               if (list_empty(&pre_select_list))
+                       return;
+               unregister_task(t);
        }
 }
 
@@ -54,18 +72,24 @@ static void sched_post_select(struct sched *s)
        struct task *t, *tmp;
 
        list_for_each_entry_safe(t, tmp, &post_select_list, post_select_node) {
-               t->post_select(s, t);
+               if (t->error >= 0)
+                       t->post_select(s, t);
 //             PARA_INFO_LOG("%s: %d\n", t->status, t->ret);
-               if (t->ret > 0 || !t->event_handler)
+               if (t->error >= 0)
                        continue;
-               t->event_handler(t);
+               /* nec., see sched_preselect() */
+               if (list_empty(&post_select_list))
+                       return;
+               unregister_task(t);
        }
 }
 
 /**
- * the core function for all paraslash programs
+ * The core function for all paraslash programs.
+ *
+ * \param s Pointer to the scheduler struct.
  *
- * Short and sweet. It updates the global \a now pointer, calls all registered
+ * This function updates the global \a now pointer, calls all registered
  * pre_select hooks which may set the timeout and add any file descriptors to
  * the fd sets of \a s.  Next, it calls para_select() and makes the result available
  * to the registered tasks by calling their post_select hook.
@@ -75,8 +99,10 @@ static void sched_post_select(struct sched *s)
  *
  * \sa task, now.
  */
-int sched(struct sched *s)
+int schedule(struct sched *s)
 {
+       int ret;
+
        if (!initialized)
                return -E_NOT_INITIALIZED;
        gettimeofday(now, NULL);
@@ -86,10 +112,11 @@ again:
        s->timeout = s->default_timeout;
        s->max_fileno = -1;
        sched_preselect(s);
-       s->select_ret = para_select(s->max_fileno + 1, &s->rfds,
-               &s->wfds, &s->timeout);
-       if (s->select_ret < 0)
-               return s->select_ret;
+       if (list_empty(&pre_select_list) && list_empty(&post_select_list))
+               return 0;
+       ret = para_select(s->max_fileno + 1, &s->rfds, &s->wfds, &s->timeout);
+       if (ret < 0)
+               return ret;
        gettimeofday(now, NULL);
        sched_post_select(s);
        if (list_empty(&pre_select_list) && list_empty(&post_select_list))
@@ -98,18 +125,18 @@ again:
 }
 
 /**
- * initialize the paraslash scheduler
+ * Initialize the paraslash scheduler.
  */
 static void init_sched(void)
 {
-       PARA_INFO_LOG("%s", "initializing scheduler\n");
+       PARA_INFO_LOG("initializing scheduler\n");
        INIT_LIST_HEAD(&pre_select_list);
        INIT_LIST_HEAD(&post_select_list);
        initialized = 1;
 };
 
 /**
- * add a task to the scheduler
+ * Add a task to the scheduler.
  *
  * \param t the task to add
  *
@@ -125,37 +152,18 @@ void register_task(struct task *t)
        PARA_INFO_LOG("registering %s (%p)\n", t->status, t);
        if (t->pre_select) {
                PARA_DEBUG_LOG("pre_select: %p\n", &t->pre_select);
-               para_list_add(&t->pre_select_node, &pre_select_list);
+               list_add_tail(&t->pre_select_node, &pre_select_list);
        }
        if (t->post_select) {
                PARA_DEBUG_LOG("post_select: %p\n", &t->pre_select);
-               para_list_add(&t->post_select_node, &post_select_list);
+               list_add_tail(&t->post_select_node, &post_select_list);
        }
 }
 
 /**
- * remove a task from the scheduler
- *
- * \param t the task to remove
- *
- * If the pre_select pointer of \a t is not \p NULL, it is removed from
- * the pre_select list of the scheduler. Same goes for \a post_select.
- */
-void unregister_task(struct task *t)
-{
-       if (!initialized)
-               return;
-       PARA_INFO_LOG("unregistering %s (%p)\n", t->status, t);
-       if (t->pre_select)
-               list_del(&t->pre_select_node);
-       if (t->post_select)
-               list_del(&t->post_select_node);
-};
-
-/**
- * unregister all tasks
+ * Unregister all tasks.
  *
* This will cause \a sched() to return immediately because both the
  * \a pre_select_list and the \a post_select_list are empty.
  */
 void sched_shutdown(void)
@@ -166,16 +174,15 @@ void sched_shutdown(void)
                return;
        list_for_each_entry_safe(t, tmp, &pre_select_list, pre_select_node)
                unregister_task(t);
-       /* remove tasks which do not have a pre_select hook */
        list_for_each_entry_safe(t, tmp, &post_select_list, post_select_node)
                unregister_task(t);
        initialized = 0;
 };
 
 /**
- * get the list of all registered tasks.
+ * Get the list of all registered tasks.
  *
- * \return the task list
+ * \return The task list.
  *
  * Each entry of the list contains an identifier which is simply a hex number
  * that may be used in \a kill_task() to terminate the task.
@@ -207,12 +214,12 @@ char *get_task_list(void)
 }
 
 /**
- * simulate an error for the given task
+ * Simulate an error for the given task.
  *
- * \param id the task identifier
+ * \param id The task identifier.
  *
* Find the task identified by \a id, set the tasks' return value to
* \p -E_TASK_KILLED and call the event handler of the task.
  *
  * \return Positive on success, negative on errors (e.g. if \a id does not
  * correspond to a registered task).
@@ -228,18 +235,14 @@ int kill_task(char *id)
                sprintf(buf, "%p", t);
                if (strcmp(id, buf))
                        continue;
-               t->ret = -E_TASK_KILLED;
-               if (t->event_handler)
-                       t->event_handler(t);
+               t->error = -E_TASK_KILLED;
                return 1;
        }
        list_for_each_entry_safe(t, tmp, &post_select_list, post_select_node) {
                sprintf(buf, "%p", t);
                if (strcmp(id, buf))
                        continue;
-               t->ret = -E_TASK_KILLED;
-               if (t->event_handler)
-                       t->event_handler(t);
+               t->error = -E_TASK_KILLED;
                return 1;
        }
        return -E_NO_SUCH_TASK;
@@ -125,37 +152,18 @@ void register_task(struct task *t)
  *
- * This will cause \a sched() to return immediately because both the
+ * This will cause \a schedule() to return immediately because both the
  * \a pre_select_list and the \a post_select_list are empty.
  */
@@ -207,12 +214,12 @@ char *get_task_list(void)
  *
- * Find the task identified by \a id, set the tasks' return value to
- * \p -E_TASK_KILLED and call the event handler of the task.
+ * Find the task identified by \a id, set the tasks' error value to
+ * \p -E_TASK_KILLED and unregister the task.
  *
  * \return Positive on success, negative on errors (e.g. if \a id does not